import gzip
import io
from typing import Optional
from xml.etree import ElementTree
from PIL import Image

import iso639
from internetarchiveocr.language import language_to_alpha3lang

#: Contains the HOCR schema
HOCR_SCHEMA = '{http://www.w3.org/1999/xhtml}'

def register_and_nuke_xhtml_namespace():
    # Nuke the namespace, otherwise it will prefix everything with html:
    ElementTree.register_namespace('', 'http://www.w3.org/1999/xhtml')


def iterparse_tags(fp, tag=None, events=None):
    doc = ElementTree.iterparse(fp, events=events)
    for act, elem in doc:
        if tag is not None and elem.tag not in tag:
            continue

        yield act, elem


def elem_tostring(elem, xml_declaration=None, short_empty_elements=False):
    s = ElementTree.tostring(elem, method='xml',
                             encoding='UTF-8',
                             short_empty_elements=short_empty_elements,
                             xml_declaration=xml_declaration)
    return s

def open_if_required(fd_or_path):
    """
    Opens a file if `fd_or_path` is a `str`, otherwise returns `fd_or_path`.
    If `fd_or_path` ends with `.gz`, uses `gzip.open`.
    """
    if isinstance(fd_or_path, str):
        if fd_or_path.endswith('.gz'):
            xml_file = gzip.open(fd_or_path, 'rb')
        else:
            xml_file = open(fd_or_path, 'rb')
    else:
        xml_file = fd_or_path

    return xml_file


def get_ocr_system(fd):
    """
    Read the ocr-system meta tag from a new file descriptor containing a hOCR
    document. If you want to use an existing file descriptor, ensure to seek to
    the start first.

    Args:

    * fd: Open file descriptor

    Return:

    * string of the content system or None if none is specified
    """
    header, footer = get_header_footer(fd)

    bio = io.BytesIO()
    bio.write(header + footer)
    bio.seek(0)

    parse = iterparse_tags(bio, tag=(HOCR_SCHEMA+'meta',), events=('end',))
    for (start_end, element) in parse:
        if element.attrib.get('name') == 'ocr-system':
            return element.attrib.get('content')

    return None


def get_header_footer(fd):
    """
    Extract the parts before and after the body elements from a given XML file

    Args:

    * fd: Open file descriptor

    Returns:

    * Tuple (header, footer)
    """
    s = ''
    tags = (HOCR_SCHEMA + 'html', HOCR_SCHEMA + 'head')
    doc = iterparse_tags(fd, tag=tags, events=('start', 'end'))
    html_elem = None
    head_elem = None

    for act, elem in doc:
        if elem.tag[-4:] == 'html' and act == 'start':
            html_elem = elem
            children = list(html_elem)
            for child in children:
                if child.tag[-4:] == 'body':
                    chs = list(child)
                    for c in chs:
                        child.remove(c)
                    # Remove body, we add an empty one
                    html_elem.remove(child)

        if elem.tag[-4:] == 'head' and act == 'end':
            head_elem = elem
            body_elem = ElementTree.Element('body')

            html_elem.append(body_elem)

            s = elem_tostring(html_elem, xml_declaration=True,
                              short_empty_elements=True)
            s = s.decode('utf-8')
            break

    comp = s.split('<body />')

    # XML-ho
    header = comp[0] + '<body>' + '\n'
    htmlidx = header.find('<html')
    doctype = '''<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">'''
    header = header[:htmlidx] + doctype + '\n' + header[htmlidx:]

    # Compatibility with previous lxml code - also Tesseract seems inconsistent
    # in this regard
    if '<title />' in header:
        header = header.replace('<title />', '<title></title>')

    footer = '</body>' + comp[1].lstrip()

    return header.encode('utf-8'), footer.encode('utf-8')


def needs_grayscale_conversion(image: Image.Image) -> bool:
    """
    Return True if image should be converted to grayscale, and False otherwise.
    Grayscale images with alpha layers need conversion because the JPEG target
    doesn't support transparency.

    Note: this will return False for RGB images that are functionally grayscale,
    as the cost of identifying them is not worth the effort. For a good, yet
    still too slow strategy, see https://stackoverflow.com/a/34175631.
    """
    return image.mode in ('LA', '1')


def needs_rgb_conversion(image: Image.Image) -> bool:
    """
    Return True if image should be converted to RGB, and False otherwise.

    Anything that isn't a grayscale image that also isn't already RGB needs conversion.
    """
    return image.mode not in ('RGB', 'L', 'LA', '1')


def normalize_language(language) -> Optional[str]:
    """
    Attempt to convert a language tag to a valid country code
    """
    if not language:
        return None

    def get_alpha2_code(language: str, part: str) -> Optional[str]:
        """
        Try to get a two-character language code from a three character code.

        `part` is an ISO 639 part (e.g. part3, part2b).
        """
        try:
            return iso639.languages.get(**{part: language_to_alpha3lang(language)}).alpha2
        except KeyError:
            return None

    alpha2_code = get_alpha2_code(language=language, part="part3")
    if alpha2_code is None:
        alpha2_code = get_alpha2_code(language=language, part="part2b")

    return alpha2_code or language



from typing import Optional

ROMAN_ARABIC_MAP = {
    "I": 1,
    "V": 5,
    "X": 10,
    "L": 50,
    "C": 100,
    "D": 500,
    "M": 1000,
}


def roman_to_num(s: str, up_to: Optional[str] = None) -> int:
    """
    Convert a Roman numeral string, s, to an int.

    Returns 0 if the input is not in `ROMAN_ARABIC_MAP`.

    Handles standard and other additive and substractive forms.
    See https://en.wikipedia.org/wiki/Roman_numerals#Other_additive_forms and
    https://en.wikipedia.org/wiki/Roman_numerals#Other_subtractive_forms.

    Courtesy of Arkiver. :)
    """
    s = s.upper()

    value = 0
    for i, (c1, c2) in enumerate(zip([s[0], *s], s)):
        if c1 not in ROMAN_ARABIC_MAP or c2 not in ROMAN_ARABIC_MAP:
            return 0
        if up_to is not None and ROMAN_ARABIC_MAP[c2] >= ROMAN_ARABIC_MAP[up_to]:
            return value
        if c1 is None or ROMAN_ARABIC_MAP[c1] >= ROMAN_ARABIC_MAP[c2]:
            value += ROMAN_ARABIC_MAP[c2]
        else:
            value += ROMAN_ARABIC_MAP[c2] - roman_to_num(s[:i][::-1], up_to=c2) * 2
    return value



import pytest
from os.path import join

def read_file(tmp_path_factory, n):
    data = tmp_path_factory.getbasetemp() / n
    with open(join('test-files', n), 'rb') as fp:
        data.write_bytes(fp.read())
    return data

@pytest.fixture(scope='session')
def sim_hocr_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'sim_english-illustrated-magazine_1884-12_2_15_chocr.html.gz')

@pytest.fixture(scope='session')
def sim_hocr_lookup_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'sim_english-illustrated-magazine_1884-12_2_15_hocr_pageindex.json')

@pytest.fixture(scope='session')
def sim_hocr_searchtext_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'sim_english-illustrated-magazine_1884-12_2_15_hocr_searchtext.txt')

@pytest.fixture(scope='session')
def pdf_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'test_arlington_heights_daily_herald-19760903.pdf')

@pytest.fixture(scope='session')
def pdf_metadata_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'test_arlington_heights_daily_herald-19760903_pdfmeta.json')

@pytest.fixture(scope='session')
def pdf_chocr_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'test_arlington_heights_daily_herald-19760903_chocr.html')

@pytest.fixture(scope='session')
def sim_english_pagenumber_json_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'sim_english_pagenumbers.json')

@pytest.fixture(scope='session')
def abbyy_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'sim_english-illustrated-magazine_1884-12_2_15_abbyy')

@pytest.fixture(scope='session')
def abbyy_to_hocr_file(tmp_path_factory):
    return read_file(tmp_path_factory, 'sim_english-abbyy-to-hocr-result.html')

@pytest.fixture(scope='session')
def sim_hocr_file_text(tmp_path_factory):
    return read_file(tmp_path_factory, 'sim_english-illustrated-magazine_1884-12_2_15_chocr.txt')



import sys
import xml.etree.ElementTree as ET
import zipfile
from datetime import datetime
from typing import Dict, List, Optional, Tuple, Union
from xml.dom import minidom

import pkg_resources

from .util import roman_to_num

# This becomes the dtb:generator meta in the generated book
content_generator = 'Internet Archive - archive.org'


# TODO: candidate for removal/refactor
def get_metadata_tag_data(metadata: List[Dict[str, str]], tag: str) -> str:
    for el in metadata:
        if el['tag'] == tag:
            return el['text']
    return 'Unknown'


# TODO: candidate for removal/refactor
def tree_to_str(tree: ET.ElementTree, xml_declaration: bool = True) -> bytes:
    return ET.tostring(
        tree.getroot(), encoding='utf-8', xml_declaration=xml_declaration
    )


class DaisyBook:
    def __init__(
        self,
        out_name: str,
        metadata: Optional[List[Dict[str, str]]] = None,
        content_dir: str = '',
    ) -> None:
        self.dt = datetime.now()
        self.z = zipfile.ZipFile(out_name, 'w')
        self.content_dir = content_dir
        self.book_id = get_metadata_tag_data(metadata, 'identifier') if metadata else ""
        self.title = get_metadata_tag_data(metadata, 'title') if metadata else ""
        self.author = get_metadata_tag_data(metadata, 'creator') if metadata else ""
        self.nav_number = 1

        self.opf_file = self.book_id + '_daisy.opf'

        self.dtbook_file = self.book_id + '_daisy.xml'
        self.dtbook, self.dtbook_book_el = make_dtbook(self.book_id, self.title)

        self.smil_file = self.book_id + '_daisy.smil'
        self.smil, self.smil_seq_el = make_smil(self.book_id)

        self.ncx_file = self.book_id + '_daisy.ncx'
        self.ncx, self.ncx_head_el, self.ncx_navmap_el, self.ncx_pagelist_el = make_ncx(
            self.book_id, self.title, self.author
        )

        self.tag_stack = [self.dtbook_book_el]
        self.navpoint_stack = [self.ncx_navmap_el]

        self.id_index = 1

        self.depth = 0
        self.current_depth = 0
        self.total_page_count = 0
        self.max_page_number = 0

        # style sheet, etc.
        for content in [
            'daisy.css',
            'daisyTransform.xsl',
            'dtbook-2005-3.dtd',
            'html.css',
            'resource.res',
        ]:
            content_src = pkg_resources.resource_filename(
                'hocr', f'daisy/daisy_files/{content}'
            )
            content_str = open(content_src).read()
            self.add(self.content_dir + content, content_str)

        self.manifest_items = [
            {
                'id': 'xml',
                'href': self.dtbook_file,
                'media-type': 'application/x-dtbook+xml',
            },
            {
                'id': 'opf',
                'href': self.book_id + '_daisy.opf',
                'media-type': 'text/xml',
            },
            {
                'id': 'ncx',
                'href': self.ncx_file,
                'media-type': 'application/x-dtbncx+xml',
            },
            {'id': 'smil', 'href': self.smil_file, 'media-type': 'application/smil'},
            {
                'id': 'daisyTransform',
                'href': 'daisyTransform.xsl',
                'media-type': 'text/xsl',
            },
            {'id': 'daisyCss', 'href': 'daisy.css', 'media-type': 'text/css'},
            {'id': 'htmlCss', 'href': 'html.css', 'media-type': 'text/css'},
            {
                'id': 'resource',
                'href': 'resource.res',
                'media-type': 'application/x-dtbresource+xml',
            },
        ]

    def push_tag(
        self,
        tag: str,
        text: str = '',
        attrs: Dict[str, str] = {},
    ) -> str:
        # tag is e.g. frontmatter, bodymatter, rearmatter, level, etc.
        id_str, dtb_el = self.add_tag(tag, text, attrs)
        self.tag_stack.append(dtb_el)
        return id_str

    def pop_tag(self) -> None:
        self.tag_stack.pop()

    def add_tag(
        self,
        tag: str,
        text: str = '',
        attrs: Dict[str, str] = {},
        smil_attrs: Dict[str, str] = {},
    ) -> Tuple[str, ET.Element]:
        id_str = tag + '_' + (str(self.id_index).zfill(5))
        attrs['id'] = id_str
        if text is not None and len(text) > 0:
            smil_attrs.update({'id': id_str, 'class': tag})
            smil_par_el = ET.SubElement(self.smil_seq_el, 'par', smil_attrs)
            ET.SubElement(
                smil_par_el,
                'text',
                {'src': self.dtbook_file + '#' + id_str, 'region': 'textRegion'},
            )
            attrs['smilref'] = self.smil_file + '#' + id_str
        current_dtb_el = self.tag_stack[-1]

        dtb_el = ET.SubElement(current_dtb_el, tag, attrs)
        if text is None:
            print(tag)

        if text is not None and len(text) > 0:
            dtb_el.text = text

        self.id_index += 1
        return id_str, dtb_el

    def add_navpoint(
        self,
        ltag: str,
        htag: str,
        text: str,
    ) -> ET.Element:
        level = str(len(self.navpoint_stack))
        level_id_str = self.push_tag(ltag + level)
        htag_id_str, htag_dtb_el = self.add_tag(
            htag + level, text, smil_attrs={'customTest': 'headerCustomTest'}
        )
        current_navpoint_el = self.navpoint_stack[-1]
        navpoint_el = ET.SubElement(
            current_navpoint_el,
            'navPoint',
            {
                'id': level_id_str,
                'class': 'navpoint-level-level' + level,
                'playOrder': str(self.nav_number),
            },
        )
        navlabel_el = ET.SubElement(navpoint_el, 'navLabel')
        ET.SubElement(navlabel_el, 'text').text = text
        ET.SubElement(
            navpoint_el, 'content', {'src': self.smil_file + '#' + htag_id_str}
        )
        self.nav_number += 1
        return navpoint_el

    def push_navpoint(self, ltag: str, htag: str, text: str) -> None:
        self.current_depth += 1
        if self.current_depth > self.depth:
            self.depth = self.current_depth
        navpoint_el = self.add_navpoint(ltag, htag, text)
        self.navpoint_stack.append(navpoint_el)

    def pop_navpoint(self) -> ET.Element:
        self.current_depth -= 1
        self.pop_tag()
        return self.navpoint_stack.pop()

    def add_pagetarget(self, name: str, value: str, type_: str = 'normal') -> None:
        self.total_page_count += 1

        if isinstance(value, int):
            int_value = value
        elif value.isdigit():
            int_value = int(value)
        elif roman_to_num(value):
            int_value = roman_to_num(value)
        else:
            return

        if int_value > self.max_page_number:
            self.max_page_number = int_value

        pagenum_id, pagenum_el = self.add_tag(
            'pagenum',
            name,
            attrs={'page': type_},
            smil_attrs={'customTest': 'pagenumCustomTest'},
        )
        pagetarget_el = ET.SubElement(
            self.ncx_pagelist_el,
            'pageTarget',
            {
                'id': pagenum_id,
                'value': str(value),
                'type': type_,
                'playOrder': str(self.nav_number),
            },
        )
        navlabel_el = ET.SubElement(pagetarget_el, 'navLabel')
        ET.SubElement(navlabel_el, 'text').text = name
        ET.SubElement(
            pagetarget_el, 'content', {'src': self.smil_file + '#' + pagenum_id}
        )
        self.nav_number += 1

    def add(
        self,
        path: str,
        content_str: Union[bytes, str],
        deflate: bool = True,
    ) -> None:
        info = zipfile.ZipInfo(path)
        info.compress_type = zipfile.ZIP_DEFLATED if deflate else zipfile.ZIP_STORED
        info.external_attr = 0o666 << 16  # fix access
        info.date_time = (
            self.dt.year,
            self.dt.month,
            self.dt.day,
            self.dt.hour,
            self.dt.minute,
            self.dt.second,
        )
        self.z.writestr(info, content_str)

    def finish(self, metadata: Optional[List[Dict[str, str]]]) -> None:
        """Do the best we can to create the opf, ncx, dtbook, smil, etc., even if metadata is None."""
        root = "<?xml version='1.0' encoding='utf-8'?>\n"

        prefix_xml = (
            root
            + '<!DOCTYPE package PUBLIC "+//ISBN 0-9673008-1-9//DTD OEB 1.2 Package//EN" "http://openebook.org/dtds/oeb-1.2/oebpkg12.dtd">'
        )
        tree_str = make_opf(metadata, self.manifest_items)
        self.add(self.content_dir + self.opf_file, tree_str)

        metas = [
            {'name': 'dtb:depth', 'content': str(self.depth)},
            {'name': 'dtb:totalPageCount', 'content': str(self.total_page_count)},
            {'name': 'dtb:maxPageNumber', 'content': str(self.max_page_number)},
        ]
        for item in metas:
            ET.SubElement(self.ncx_head_el, 'meta', item)
        prefix_xml = (
            root
            + '<!DOCTYPE ncx PUBLIC "-//NISO//DTD ncx 2005-1//EN" "http://www.daisy.org/z3986/2005/ncx-2005-1.dtd">'
        )
        tree_str = tree_to_str_with_processing(self.ncx, prefix_xml)
        self.add(self.content_dir + self.ncx_file, tree_str)

        doctype = '<!DOCTYPE dtbook SYSTEM "dtbook-2005-3.dtd">\n'
        pi1 = '<?xml-stylesheet type="text/css" href="daisy.css" media="screen"?>\n'
        pi2 = '<?xml-stylesheet type="text/xsl" href="daisyTransform.xsl" media="screen"?>\n'
        prefix_xml = root + doctype + pi1 + pi2
        tree_str = tree_to_str_with_processing(self.dtbook, prefix_xml)
        self.add(self.content_dir + self.dtbook_file, tree_str)

        prefix_xml = (
            root
            + '<!DOCTYPE smil PUBLIC "-//NISO//DTD dtbsmil 2005-2//EN" "http://www.daisy.org/z3986/2005/dtbsmil-2005-2.dtd">\n'
        )
        tree_str = tree_to_str_with_processing(self.smil, prefix_xml)
        self.add(self.content_dir + self.smil_file, tree_str)

        self.z.close()


dc = 'http://purl.org/dc/elements/1.1/'
dcb = '{' + dc + '}'


def make_opf(
    metadata: Optional[List[Dict[str, str]]], manifest_items: List[Dict[str, str]]
) -> str:
    # If no metadata is provided, just create a skeleton file.
    # Note: it may be possible to get the title from the XML.
    if not metadata:
        metadata = [{}]

    root_el = ET.Element(
        'package',
        {
            'xmlns': 'http://openebook.org/namespaces/oeb-package/1.0/',
            'unique-identifier': 'bookid',
        },
    )

    metadata_el = ET.SubElement(root_el, 'metadata')
    dc_metadata_el = ET.SubElement(
        metadata_el,
        'dc-metadata',
        {
            'xmlns:dc': 'http://purl.org/dc/elements/1.1/',
            'xmlns:oebpackage': 'http://openebook.org/namespaces/oeb-package/1.0/',
        },
    )

    el = ET.SubElement(dc_metadata_el, 'dc:Format')
    el.text = 'ANSI/NISO Z39.86-2005'

    for md in metadata:
        tagname = md.get('tag')
        if tagname not in [
            'title',
            'creator',
            'subject',
            'description',
            'publisher',
            'contributor',
            'date',
            'type',
            'format',
            'identifier',
            'source',
            'language',
            'relation',
            'coverage',
            'rights',
        ]:
            continue
        dctag = f'dc:{tagname[:1].upper()}{tagname[1:]}'
        if tagname == 'identifier':
            el = ET.SubElement(dc_metadata_el, dctag, {'id': 'bookid'})
        else:
            el = ET.SubElement(dc_metadata_el, dctag)
        el.text = md['text']

    x_metadata_el = ET.SubElement(metadata_el, 'x-metadata')
    ET.SubElement(
        x_metadata_el, 'meta', {'name': 'dtb:multimediaType', 'content': 'textNCX'}
    )
    ET.SubElement(
        x_metadata_el, 'meta', {'name': 'dtb:multimediaContent', 'content': 'text'}
    )
    ET.SubElement(x_metadata_el, 'meta', {'name': 'dtb:totalTime', 'content': '0'})

    manifest_el = ET.SubElement(root_el, 'manifest')
    for item in manifest_items:
        ET.SubElement(manifest_el, 'item', item)

    spine_el = ET.SubElement(root_el, 'spine')
    ET.SubElement(spine_el, 'itemref', {'idref': 'smil'})

    tree = ET.ElementTree(root_el)
    root = "<?xml version='1.0' encoding='utf-8'?>\n"
    prefix_xml = (
        root
        + '<!DOCTYPE package PUBLIC "+//ISBN 0-9673008-1-9//DTD OEB 1.2 Package//EN" "http://openebook.org/dtds/oeb-1.2/oebpkg12.dtd">'
    )
    return tree_to_str_with_processing(tree, prefix_xml)


def tree_to_str_with_processing(tree: ET.ElementTree, prefix_xml: str) -> str:
    xml_str = ET.tostring(
        tree.getroot(),
        encoding='unicode',
        method='xml',
    )
    return minidom.parseString(prefix_xml + xml_str).toprettyxml(indent="  ")


def make_dtbook(book_id: str, title: str) -> Tuple[ET.ElementTree, ET.Element]:
    root_el = ET.Element(
        'dtbook',
        {'xmlns': 'http://www.daisy.org/z3986/2005/dtbook/', 'version': '2005-3'},
    )

    title = title if title else book_id

    head_el = ET.SubElement(root_el, 'head')
    ET.SubElement(head_el, 'meta', {'name': 'dtb:uid', 'content': book_id})
    ET.SubElement(head_el, 'meta', {'name': 'dc:Title', 'content': title})
    book_el = ET.SubElement(root_el, 'book')

    tree = ET.ElementTree(root_el)
    return tree, book_el


def make_smil(book_id: str) -> Tuple[ET.ElementTree, ET.Element]:
    root_el = ET.Element('smil', {'xmlns': 'http://www.w3.org/2001/SMIL20/'})

    head_el = ET.SubElement(root_el, 'head')
    ET.SubElement(head_el, 'meta', {'name': 'dtb:uid', 'content': book_id})
    ET.SubElement(
        head_el, 'meta', {'name': 'dtb:generator', 'content': content_generator}
    )
    ET.SubElement(head_el, 'meta', {'name': 'dtb:totalElapsedTime', 'content': '0'})
    layout_el = ET.SubElement(head_el, 'layout')
    ET.SubElement(
        layout_el,
        'region',
        {
            'id': 'textRegion',
            'fit': 'hidden',
            'showBackground': 'always',
            'height': 'auto',
            'width': 'auto',
            'bottom': 'auto',
            'top': 'auto',
            'left': 'auto',
            'right': 'auto',
        },
    )
    customattributes_el = ET.SubElement(head_el, 'customAttributes')
    ET.SubElement(
        customattributes_el,
        'customTest',
        {'id': 'pagenumCustomTest', 'defaultState': 'false', 'override': 'visible'},
    )
    ET.SubElement(
        customattributes_el,
        'customTest',
        {'id': 'headerCustomTest', 'defaultState': 'false', 'override': 'visible'},
    )

    body_el = ET.SubElement(root_el, 'body')
    seq_el = ET.SubElement(body_el, 'seq', {'id': 'toplevel_seq_id'})
    tree = ET.ElementTree(root_el)
    return tree, seq_el


def make_ncx(
    book_id: str, title: str, author: str
) -> Tuple[ET.ElementTree, ET.Element, ET.Element, ET.Element]:
    root_el = ET.Element(
        'ncx', {'xmlns': 'http://www.daisy.org/z3986/2005/ncx/', 'version': '2005-1'}
    )

    head_el = ET.SubElement(root_el, 'head')
    ET.SubElement(
        head_el,
        'smilCustomTest',
        {
            'id': 'pagenumCustomTest',
            'defaultState': 'false',
            'override': 'visible',
            'bookStruct': 'PAGE_NUMBER',
        },
    )
    ET.SubElement(
        head_el,
        'smilCustomTest',
        {
            'id': 'headerCustomTest',
            'defaultState': 'false',
            'override': 'visible',
            'bookStruct': 'PAGE_NUMBER',
        },
    )
    metas = [
        {'name': 'dtb:uid', 'content': book_id},
        {'name': 'dtb:generator', 'content': content_generator},
    ]
    for item in metas:
        ET.SubElement(head_el, 'meta', item)

    doctitle = ET.SubElement(root_el, 'docTitle')
    ET.SubElement(doctitle, 'text').text = title
    docauthor = ET.SubElement(root_el, 'docAuthor')
    ET.SubElement(docauthor, 'text').text = author

    navmap_el = ET.SubElement(root_el, 'navMap')
    navinfo_el = ET.SubElement(navmap_el, 'navInfo')
    ET.SubElement(navinfo_el, 'text').text = 'Book navigation'

    pagelist_el = ET.SubElement(root_el, 'pageList')
    navlabel_el = ET.SubElement(pagelist_el, 'navLabel')
    ET.SubElement(navlabel_el, 'text').text = 'Pages'

    tree = ET.ElementTree(root_el)
    return tree, head_el, navmap_el, pagelist_el


if __name__ == '__main__':
    sys.stderr.write('I\'m a module.  Don\'t run me directly!')



from hocr.parse import hocr_page_to_word_data_fast, hocr_page_get_dimensions
from hocr.searching import hocr_lookup_page_by_dat, \
        hocr_lookup_by_plaintext_offset
from hocr.text import get_paragraph_hocr_words, hocr_paragraph_text, \
        get_paragraph_hocr_words, hocr_page_text_from_word_data

"""
Highly experimental and unstable interface to retrieve page indexes and
bounding boxes for words matching a certain full-text-search query.

Required an external system to perform the highlighting (bin/fts-text-annotate
can serve this purpose).

This API is highly unstable, the code is in need of cleanups, and there likely
isn't a big use for this particular file outside of Internet Archive purposes,
unless you want to match ElasticSearch highlighted plaintext to hOCR pages to
get the text coordinates of the matches.
"""


FINAL_PRE_TAG = '{{{'
FINAL_POST_TAG = '}}}'

def match_words(hocr_words, match_indexes):
    def m(match, word_start, word_end):
        # XXX: We might need to turn some of these <= into < or so, there could
        # be an off-by-one error here, matching a next word when it should not.
        return (match[0] >= word_start and match[0] <= word_end) or \
               (match[1] >= word_start and match[1] <= word_end) or \
               (word_start >= match[0] and word_end <= match[0]) or \
               (word_start >= match[1] and word_end <= match[1])

    matching_words = []
    in_word = False
    current_match_words = []

    current_match_index = 0
    str_idx = 0

    for word in hocr_words:
        if current_match_index >= len(match_indexes):
            break

        word_start = str_idx
        word_end = str_idx + len(word['text'])

        if in_word:
            # match in the word, or word in match
            if m(match_indexes[current_match_index], word_start, word_end):
                current_match_words.append(word)
            else:
                matching_words.append(current_match_words)
                current_match_words = []
                current_match_index += 1
                in_word = False

        if not in_word and current_match_index < len(match_indexes):
            if m(match_indexes[current_match_index], word_start, word_end):
                current_match_words.append(word)

                in_word = True

        # Add + 1 for the space after a word
        str_idx += len(word['text']) + 1

    if current_match_index < len(match_indexes):
        word_start = str_idx
        word_end = str_idx + len(word['text'])

        if in_word:
            if m(match_indexes[current_match_index], word_start, word_end):
                current_match_words.append(word)
            else:
                matching_words.append(current_match_words)
                current_match_words = []
                current_match_index += 1
                in_word = False
        else:
            if m(match_indexes[current_match_index], word_start, word_end):
                current_match_words.append(word)
                matching_words.append(current_match_words)

    return matching_words


def find_word_boxes(solr_line, hocr_text, hocr_par, page, page_no, pre_tag, post_tag, replace_with_final_tags=False):
    match_number = 0
    match_with = solr_line
    cur = {
        'text': solr_line,
        'par': []
    }

    results = []

    # Contains a tuple for each match, with the starting and ending rune
    match_indexes = []

    # Match solr_line words to hocr_par words
    sub_idx = 0
    while True:
        s = solr_line[sub_idx:].find(pre_tag)
        if s == -1:
            break
        rs = s + sub_idx + len(pre_tag)

        sub_idx += s
        e = solr_line[sub_idx:].find(post_tag)
        if e == -1:
            break
        re = e + sub_idx
        sub_idx += e

        match_indexes.append((rs, re))

    # Normalise indices for string without pre_tag and post_tag, this makes life easier
    # later on
    for idx in range(len(match_indexes)):
        rs, re = match_indexes[idx]
        match_indexes[idx] = (rs - (len(pre_tag) + len(post_tag)) * (idx) - len(pre_tag),
                              re - (len(pre_tag) + len(post_tag)) * (idx) - len(post_tag))

    if not len(match_indexes):
        if solr_line.find(post_tag) < solr_line.find(pre_tag):
            # XXX: Known bug, this happens because elastic currently matches
            # across paragraph boundaries, which we do not support. We could
            # extend our match to support multi-paragraph matching, but we're
            # probably going to change our elastic search to not match across
            # paragraphs or even pages.
            pass
        else:
            raise Exception('No match_indexes for %s' % repr(solr_line))

        return None

    # Now we know where our matches are in the text (without the brackets), we
    # need to match that to the hocr words, because just knowing which text is
    # not relevant - we need to find the bounding boxes for the matching text.
    hocr_words = get_paragraph_hocr_words(hocr_par)
    word_matches = match_words(hocr_words, match_indexes)

    # We have bounding boxes per word, but the current API doesn't permit us to
    # pass multiple bounding boxes for a single match, so let's find the
    # encompassing bounding box
    boxes = []
    for words in word_matches:
        boxes.append({
            'l': min([x['bbox'][0] for x in words]),
            't': min([x['bbox'][1] for x in words]),
            'r': max([x['bbox'][2] for x in words]),
            'b': max([x['bbox'][3] for x in words]),
            'page': page_no,
        })

    left = None
    top = None
    right = None
    bottom = None
    for box in boxes:
        if left is None:
            left = box['l']
        else:
            left = min(left, box['l'])

        if top is None:
            top = box['t']
        else:
            top = min(top, box['t'])

        if right is None:
            right = box['r']
        else:
            right = max(right, box['r'])

        if bottom is None:
            bottom = box['b']
        else:
            bottom = max(bottom, box['b'])
    all_boxes = {'l': left, 't': top, 'r': right, 'b': bottom}

    r = all_boxes
    page_width, page_height = hocr_page_get_dimensions(page)
    r.update({'page': page_no,
              'boxes': boxes,
              'page_width': page_width,
              'page_height': page_height})

    cur['par'].append(r)

    if replace_with_final_tags:
        # Whatever the tag was, substitute it back for {{{ and }}}
        if pre_tag != FINAL_POST_TAG:
            cur['text'] = cur['text'].replace(pre_tag, FINAL_PRE_TAG)
        if post_tag != FINAL_PRE_TAG:
            cur['text'] = cur['text'].replace(post_tag, FINAL_POST_TAG)

    results.append(cur)
    return results


def find_matches(lookup_table, hocrfp, text, es_whitespace_fixup_required=False,
                 pre_tag='{{{', post_tag='}}}',
                 replace_with_final_tags=False):
    text_byte_count = 0
    current_dat = None
    page_number = 0

    if es_whitespace_fixup_required:
        if not text.endswith('\n'):
            text += '\n'

        # There might be faster ways of doing this (e.g. read the _searchtext
        # file and count the amount of 'whitespace' bytes)
        done = False
        for dat in lookup_table:
            page = hocr_lookup_page_by_dat(hocrfp, dat)
            word_data = hocr_page_to_word_data_fast(page)
            page_text = hocr_page_text_from_word_data(word_data)

            for line in page_text:
                if line.strip() == '':
                    # Add counted bytes, one for newline
                    text_byte_count += len(line)
                    continue
                else:
                    done = True
                    break
            if done:
                break


    # For every line in the highlighted text, let's find matches...
    for line in text[:-1].split('\n'):
        # Line should contain both pre_tag and post_tag
        contains_left_match = pre_tag in line
        contains_right_match = post_tag in line

        contains_match = contains_left_match and contains_right_match

        if contains_left_match or contains_right_match and not contains_match:
            # Matches span multiple lines...
            pass

        if contains_match:
            page_number, new_dat = \
                    hocr_lookup_by_plaintext_offset(lookup_table,
                                                    text_byte_count)

            # Check if we need to change/reload our page and paragraphs
            # variables
            if new_dat != current_dat:
                # Only do this if we're on a new page
                current_dat = new_dat
                page = hocr_lookup_page_by_dat(hocrfp, current_dat)
                paragraphs = hocr_page_to_word_data_fast(page)

            # Figure out what paragraph we are at, based on text length?
            # Find paragraph that contains this line, we know where the line
            # starts, so now we just need to find the paragraph, we can do this
            # with getting the paragraph text on a page, and add the amount of
            # characters, until we reach the line start.
            page_start_at = current_dat[0]
            match_at = text_byte_count
            cnt = 0
            match = None
            for idx, paragraph in enumerate(paragraphs):
                txt = hocr_paragraph_text(paragraph)
                # Add + 1 for newline
                cnt += len(txt) + 1

                if page_start_at + cnt > match_at:
                    match = idx
                    break

            if match is None:
                # This should never happen
                raise Exception('Could not find any match!')

            paragraph_words = paragraphs[match]
            paragraph_txt = hocr_paragraph_text(paragraph_words)

            # TODO: We might want to remove this in the future, it's wasteful
            # to do the replace again.
            if paragraph_txt != line.replace(pre_tag, '').replace(post_tag, ''):
                raise Exception('Reconstructed text does not match')

            word_results = find_word_boxes(line, paragraph_txt,
                                           paragraph_words, page, page_number,
                                           pre_tag, post_tag,
                                           replace_with_final_tags=replace_with_final_tags)

            # We currently (rarely) allow word_results to be empty.
            # This happens for example in a paragraph like this:
            # " THE}}} CAMP OF {{{THE BRITISH MISSION AT ADOWA. From a Drawing by F. VILLIERs.  the}}}"
            # where elastic finds very strange matches and also across
            # paragraph, which we do not support. this is also special cased in
            # find_word_boxes
            if word_results is not None:
                yield word_results

        # Correct for any {{{ and }}}, which are not in our plaintext, but are
        # in the FTS text
        subtract = 0
        if contains_left_match:
            subtract += line.count(pre_tag) * len(pre_tag)
        if contains_right_match:
            subtract += line.count(post_tag) * len(post_tag)

        # Add counted bytes, plus one for the newline
        text_byte_count += len(line) - subtract + 1
